@@ -61,7 +61,6 @@ class BlobNotFound(ArtifactError):
 #
 # Args:
 #     context (Context): The BuildStream context
-#     enable_push (bool): Whether pushing is allowed by the platform
 #
 # Pushing is explicitly disabled by the platform in some cases,
 # like when we are falling back to functioning without using
@@ -69,7 +68,7 @@ class BlobNotFound(ArtifactError):
 #
 class CASCache(ArtifactCache):
 
-    def __init__(self, context, *, enable_push=True):
+    def __init__(self, context):
         super().__init__(context)
 
         self.casdir = os.path.join(context.artifactdir, 'cas')
@@ -78,8 +77,6 @@ class CASCache(ArtifactCache):
 
         self._calculate_cache_quota()
 
-        self._enable_push = enable_push
-
         # Per-project list of _CASRemote instances.
         self._remotes = {}
 
@@ -90,6 +87,12 @@ class CASCache(ArtifactCache):
     #     Implementation of abstract methods       #
     ################################################
 
+    def preflight(self):
+        if (not os.path.isdir(os.path.join(self.casdir, 'refs', 'heads')) or
+            not os.path.isdir(os.path.join(self.casdir, 'objects'))):
+            raise ArtifactError("CAS repository check failed for '{}'"
+                                .format(self.casdir))
+
     def contains(self, element, key):
         refpath = self._refpath(self.get_artifact_fullname(element, key))
 
@@ -221,7 +224,7 @@ class CASCache(ArtifactCache):
             return bool(remotes_for_project)
 
     def has_push_remotes(self, *, element=None):
-        if not self._has_push_remotes or not self._enable_push:
+        if not self._has_push_remotes:
             # No project has push remotes
             return False
         elif element is None:
@@ -234,13 +237,13 @@ class CASCache(ArtifactCache):
 
     def pull(self, element, key, *, progress=None):
         ref = self.get_artifact_fullname(element, key)
+        display_key = key[:self.context.log_key_length]
 
         project = element._get_project()
 
         for remote in self._remotes[project]:
             try:
                 remote.init()
-                display_key = element._get_brief_display_key()
                 element.status("Pulling artifact {} <- {}".format(display_key, remote.spec.url))
 
                 request = buildstream_pb2.GetReferenceRequest()
@@ -262,14 +265,14 @@ class CASCache(ArtifactCache):
             except grpc.RpcError as e:
                 if e.code() != grpc.StatusCode.NOT_FOUND:
                     raise ArtifactError("Failed to pull artifact {}: {}".format(
-                        element._get_brief_display_key(), e)) from e
+                        display_key, e)) from e
                 else:
                     element.info("Remote ({}) does not have {} cached".format(
-                        remote.spec.url, element._get_brief_display_key()
+                        remote.spec.url, display_key
                     ))
             except BlobNotFound as e:
                 element.info("Remote ({}) does not have {} cached".format(
-                    remote.spec.url, element._get_brief_display_key()
+                    remote.spec.url, display_key
                 ))
 
         return False
