@@ -28,11 +28,10 @@ import tarfile
 from contextlib import contextmanager
 from tempfile import TemporaryDirectory
 
-from ._exceptions import StreamError, ImplError, BstError, set_last_task_error
+from ._exceptions import StreamError, ImplError, BstError
 from ._message import Message, MessageType
 from ._scheduler import Scheduler, SchedStatus, TrackQueue, FetchQueue, BuildQueue, PullQueue, PushQueue
 from ._pipeline import Pipeline, PipelineSelection
-from ._platform import Platform
 from . import utils, _yaml, _site
 from . import Scope, Consistency
 
@@ -71,8 +70,7 @@ class Stream():
         #
         # Private members
         #
-        self._platform = Platform.get_platform()
-        self._artifacts = self._platform.artifactcache
+        self._artifacts = context.artifactcache
         self._context = context
         self._project = project
         self._pipeline = Pipeline(context, project, self._artifacts)
@@ -346,6 +344,20 @@ class Stream():
         if not self._artifacts.has_push_remotes():
             raise StreamError("No artifact caches available for pushing artifacts")
 
+        # Mark all dependencies of all selected elements as "pulled" before
+        # trying to push.
+        #
+        # In non-strict mode, elements which are cached by their weak keys
+        # will attempt to pull a remote artifact by it's strict key and prefer
+        # a strict key artifact, however pull does not occur when running
+        # a `bst push` session.
+        #
+        # Marking the elements as pulled is a workaround which ensures that
+        # the cache keys are resolved before pushing.
+        #
+        for element in elements:
+            element._pull_done()
+
         self._pipeline.assert_consistent(elements)
         self._add_queue(PushQueue(self._scheduler))
         self._enqueue_plan(elements)
@@ -464,7 +476,7 @@ class Stream():
                                               selection=PipelineSelection.REDIRECT,
                                               track_selection=PipelineSelection.REDIRECT)
         target = elements[0]
-        workdir = os.path.abspath(directory)
+        directory = os.path.abspath(directory)
 
         if not list(target.sources()):
             build_depends = [x.name for x in target.dependencies(Scope.BUILD, recurse=False)]
@@ -480,7 +492,7 @@ class Stream():
         workspace = workspaces.get_workspace(target._get_full_name())
         if workspace and not force:
             raise StreamError("Workspace '{}' is already defined at: {}"
-                              .format(target.name, workspace.path))
+                              .format(target.name, workspace.get_absolute_path()))
 
         # If we're going to checkout, we need at least a fetch,
         # if we were asked to track first, we're going to fetch anyway.
@@ -506,7 +518,7 @@ class Stream():
         except OSError as e:
             raise StreamError("Failed to create workspace directory: {}".format(e)) from e
 
-        workspaces.create_workspace(target._get_full_name(), workdir)
+        workspaces.create_workspace(target._get_full_name(), directory)
 
         if not no_checkout:
             with target.timed_activity("Staging sources to {}".format(directory)):
@@ -530,12 +542,12 @@ class Stream():
         # Remove workspace directory if prompted
         if remove_dir:
             with self._context.timed_activity("Removing workspace directory {}"
-                                              .format(workspace.path)):
+                                              .format(workspace.get_absolute_path())):
                 try:
-                    shutil.rmtree(workspace.path)
+                    shutil.rmtree(workspace.get_absolute_path())
                 except OSError as e:
                     raise StreamError("Could not remove  '{}': {}"
-                                      .format(workspace.path, e)) from e
+                                      .format(workspace.get_absolute_path(), e)) from e
 
         # Delete the workspace and save the configuration
         workspaces.delete_workspace(element_name)
@@ -578,28 +590,30 @@ class Stream():
 
         for element in elements:
             workspace = workspaces.get_workspace(element._get_full_name())
-
+            workspace_path = workspace.get_absolute_path()
             if soft:
                 workspace.prepared = False
                 self._message(MessageType.INFO, "Reset workspace state for {} at: {}"
-                              .format(element.name, workspace.path))
+                              .format(element.name, workspace_path))
                 continue
 
             with element.timed_activity("Removing workspace directory {}"
-                                        .format(workspace.path)):
+                                        .format(workspace_path)):
                 try:
-                    shutil.rmtree(workspace.path)
+                    shutil.rmtree(workspace_path)
                 except OSError as e:
                     raise StreamError("Could not remove  '{}': {}"
-                                      .format(workspace.path, e)) from e
+                                      .format(workspace_path, e)) from e
 
             workspaces.delete_workspace(element._get_full_name())
-            workspaces.create_workspace(element._get_full_name(), workspace.path)
+            workspaces.create_workspace(element._get_full_name(), workspace_path)
 
-            with element.timed_activity("Staging sources to {}".format(workspace.path)):
+            with element.timed_activity("Staging sources to {}".format(workspace_path)):
                 element._open_workspace()
 
-            self._message(MessageType.INFO, "Reset workspace for {} at: {}".format(element.name, workspace.path))
+            self._message(MessageType.INFO,
+                          "Reset workspace for {} at: {}".format(element.name,
+                                                                 workspace_path))
 
         workspaces.save_config()
 
@@ -636,7 +650,7 @@ class Stream():
         for element_name, workspace_ in self._context.get_workspaces().list():
             workspace_detail = {
                 'element': element_name,
-                'directory': workspace_.path,
+                'directory': workspace_.get_absolute_path(),
             }
             workspaces.append(workspace_detail)
 
@@ -1007,17 +1021,6 @@ class Stream():
 
         _, status = self._scheduler.run(self.queues)
 
-        # Force update element states after a run, such that the summary
-        # is more coherent
-        try:
-            for element in self.total_elements:
-                element._update_state()
-        except BstError as e:
-            self._message(MessageType.ERROR, "Error resolving final state", detail=str(e))
-            set_last_task_error(e.domain, e.reason)
-        except Exception as e:   # pylint: disable=broad-except
-            self._message(MessageType.BUG, "Unhandled exception while resolving final state", detail=str(e))
-
         if status == SchedStatus.ERROR:
             raise StreamError()
         elif status == SchedStatus.TERMINATED:
